# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import struct
from typing import BinaryIO, Dict, Generator, Optional, Tuple, Union

from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib import EPath

IJSONL_SUFFIX = ".jsonl.idx"


@edataclass
class IJsonlSamplePointer:
    """
    Points to a sample inside some jsonl file on disk.
    """

    # The index of the sample in the jsonl file.
    index: int

    # The byte offset of the sample in the jsonl file.
    byte_offset: int
    # The size of the sample in the jsonl file.
    byte_size: int


class IJsonlIndexReader:
    def __init__(self, jsonl_path: Union[EPath, str]):
        jsonl_path = EPath(jsonl_path)
        index_path = jsonl_path.with_suffix(IJSONL_SUFFIX)
        self._length = index_path.size() // 8
        self.ijsonl = index_path.open("rb")

    def __getitem__(self, index: int) -> int:
        if index >= self._length or index < 0:
            raise IndexError(f"Index {index} out of range")

        if self.ijsonl.tell() != 8 * index:
            self.ijsonl.seek(8 * index)

        return struct.unpack("Q", self.ijsonl.read(8))[0]

    def __iter__(self) -> Generator[int, None, None]:
        self.ijsonl.seek(0)
        while True:
            raw = self.ijsonl.read(8)
            if len(raw) == 0:
                break
            assert len(raw) == 8
            yield struct.unpack("Q", raw)[0]

    def __len__(self) -> int:
        return self._length

    def close(self):
        self.ijsonl.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @staticmethod
    def count_samples(jsonl_path: EPath | str) -> int:
        return EPath(jsonl_path).with_suffix(IJSONL_SUFFIX).size() // 8 - 1

    @staticmethod
    def size(jsonl_path: EPath) -> int:
        with IJsonlIndexReader(jsonl_path) as reader:
            return reader[len(reader) - 1]


class IJsonlIndexWriter:
    def __init__(self, jsonl_path: EPath):
        self.final_name = jsonl_path.with_suffix(IJSONL_SUFFIX)
        self.tmp_name = jsonl_path.with_suffix(IJSONL_SUFFIX + ".tmp")
        self.ijsonl = self.tmp_name.open("wb")

    def append(self, offset: int):
        self.ijsonl.write(struct.pack("Q", offset))

    def close(self, finalize: bool = True):
        self.ijsonl.close()
        if finalize:
            self.tmp_name.move(self.final_name)
        else:
            self.tmp_name.unlink()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close(finalize=exc_val is None)


@edataclass
class CacheEntry:
    ijsonl_index_reader: IJsonlIndexReader
    lookahead_offset: Optional[int] = None
    lookahead_byteoffset: Optional[int] = None


class CachedIJsonlOffsetReader:
    """
    This class is a high-level wrapper around IJsonlIndexReader that caches some
    of the recent lookups for faster access. It is designed for the case when
    you need to read multiple offsets from the same jsonl file.

    Args:
        cache_size: The number of entries to keep in the cache. By default, we keep 32.
    """

    def __init__(self, jsonl_file: Union[str, EPath], cache_size: int = 32):
        # Maps current_offset -> CacheEntry
        self.ijsonl_index_reader_cache: Dict[int, CacheEntry] = {}
        self.cache_size = cache_size
        self.jsonl_file = EPath(jsonl_file)

    def close(self):
        for cache_entry in self.ijsonl_index_reader_cache.values():
            cache_entry.ijsonl_index_reader.close()
        self.ijsonl_index_reader_cache.clear()

    def _find_or_create_entry(
        self,
        sample_offset: int,
    ) -> CacheEntry:
        """
        1. If we already have a key == sample_offset, return it.
        2. Otherwise, create a new entry or reuse the oldest entry.
        """
        # Direct hit in the cache?
        if sample_offset in self.ijsonl_index_reader_cache:
            return self.ijsonl_index_reader_cache[sample_offset]

        # We didn't find an existing entry. Create a new one.
        # Evict if needed.
        if len(self.ijsonl_index_reader_cache) >= self.cache_size:
            # Reuse the oldest entry
            oldest_key = next(iter(self.ijsonl_index_reader_cache))
            cache_entry = self.ijsonl_index_reader_cache.pop(oldest_key)
        else:
            new_reader = IJsonlIndexReader(self.jsonl_file)
            cache_entry = CacheEntry(ijsonl_index_reader=new_reader)
        self.ijsonl_index_reader_cache[sample_offset] = cache_entry
        return cache_entry

    def _get_ijsonl_byte_offset_with_entry(
        self,
        cache_entry: CacheEntry,
        sample_offset: int,
    ) -> Tuple[int, int]:
        """
        Return (start_byte_offset, length_to_next),
        possibly using per-entry lookahead for speed.
        """
        ijsonl_index_reader = cache_entry.ijsonl_index_reader

        # If offset=0, define the result as byte offset=0 for convenience
        if sample_offset == 0:
            result_byte_offset = 0
        elif sample_offset == cache_entry.lookahead_offset:
            # Reuse the previously cached byte offset from the lookahead
            assert cache_entry.lookahead_byteoffset is not None, (
                "Lookahead offset matched but no lookahead byte offset found."
            )
            result_byte_offset = cache_entry.lookahead_byteoffset
        else:
            # Normal random access
            result_byte_offset = ijsonl_index_reader[sample_offset]

        # Prepare the lookahead for (sample_offset+1)
        next_offset = sample_offset + 1
        try:
            cache_entry.lookahead_byteoffset = ijsonl_index_reader[next_offset]
            cache_entry.lookahead_offset = next_offset
        except IndexError:
            cache_entry.lookahead_offset = None
            cache_entry.lookahead_byteoffset = None

        # length = difference to the next offset, or 0 if none
        if cache_entry.lookahead_byteoffset is not None:
            length = cache_entry.lookahead_byteoffset - result_byte_offset
        else:
            length = 0

        return result_byte_offset, length

    def get_ijsonl_byte_offset(
        self,
        sample_offset: int = 0,
    ) -> Tuple[int, int]:
        """
        High-level API to get the byte offset and length for the given file & sample_offset.
        """

        # Find or create the suitable CacheEntry
        entry = self._find_or_create_entry(sample_offset)

        # Use (and update) the per-entry lookahead logic
        result_byte_offset, length = self._get_ijsonl_byte_offset_with_entry(entry, sample_offset)

        # Update cache entry with the new offset
        self.ijsonl_index_reader_cache.pop(sample_offset)
        if entry.lookahead_offset is not None:
            new_key = entry.lookahead_offset
            if new_key not in self.ijsonl_index_reader_cache:
                self.ijsonl_index_reader_cache[new_key] = entry
            else:
                # Already have this entry in the cache, so we can close the reader and use the existing one
                # TODO: We may actually may want to keep multiple readers open, because they may be multiple
                # sequences to the same sequence.
                entry.ijsonl_index_reader.close()
        else:
            # No lookahead, so we can close the reader
            entry.ijsonl_index_reader.close()

        return result_byte_offset, length

    def __len__(self) -> int:
        if len(self.ijsonl_index_reader_cache) == 0:
            return IJsonlIndexReader.count_samples(self.jsonl_file)
        return len(next(iter(self.ijsonl_index_reader_cache.values())).ijsonl_index_reader) - 1

    def get_total_size(self) -> int:
        if len(self.ijsonl_index_reader_cache) == 0:
            self.ijsonl_index_reader_cache[0] = CacheEntry(
                ijsonl_index_reader=IJsonlIndexReader(self.jsonl_file)
            )
        reader = next(iter(self.ijsonl_index_reader_cache.values())).ijsonl_index_reader
        return reader[len(reader) - 1]


class IJsonlFile:
    """
    This class is a high-level wrapper around a binary file that allows for reading a jsonl file,
    with random access while keeping the file open.

    Usage:
        with open(filename, "rb") as fileobj:
            with IJsonlFile(fileobj=fileobj) as f:
                data = f.next(offset=101888, size=100)
                json.loads(data)
        # Or, if you want to read the whole file:
        with open(filename, "rb") as fileobj:
            with IJsonlFile(fileobj=fileobj) as f:
                while True:
                    data = f.next()
                    if data is None:
                        break
                    json.loads(data)
        # Or, if you want to read the whole file:
        with open(filename, "rb") as fileobj:
            with IJsonlFile(fileobj=fileobj) as f:
                for data in f:
                    json.loads(data)
    """

    def __init__(self, fileobj: BinaryIO):
        self.fileobj = fileobj

    def seek(self, offset: int):
        self.fileobj.seek(offset)

    def next(self, offset: int | None = None, size: int | None = None) -> bytes | None:
        if offset is not None and offset != self.fileobj.tell():
            self.fileobj.seek(offset)
        if size is None:
            entry = self.fileobj.readline()
            if entry == b"":
                return None
            return entry
        else:
            assert size > 0, "Size must contain at least the line terminator and a json object"
            data = self.fileobj.read(size)
            if data == b"":
                return None
            return data

    def __iter__(self) -> Generator[bytes, None, None]:
        while True:
            data = self.next()
            if data is None:
                break
            yield data

    def close(self):
        self.fileobj.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.fileobj.close()



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

try:
    # Try importing optional dependencies
    import av
    import av.container

except ImportError:
    pass


def av_open(file: str) -> "av.container.InputContainer":
    """Open a file with PyAV.

    This function is a wrapper around av.open that disables additional threads in the container.
    """

    input_container = av.open(file, "r")
    try:
        initialize_av_container(input_container)
    except Exception:
        input_container.close()
        raise
    return input_container


def initialize_av_container(input_container: "av.container.InputContainer") -> None:
    """Every PyAV container should be initialized with this function.

    This function ensures that no additional threads are created.
    This is to avoid deadlocks in ffmpeg when deallocating the container.
    Furthermore, we cannot have multiple threads before forking the process when
    using torch data loaders with multiple workers.
    """

    for stream in input_container.streams:
        cc = stream.codec_context

        if cc is not None:
            cc.thread_type = "NONE"
            cc.thread_count = 0



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from typing import (
    Callable,
    Generator,
    Optional,
    Tuple,
    TypeVar,
)

from megatron.energon.epathlib import EPath
from megatron.energon.flavors.jsonl.ijsonl import (
    CachedIJsonlOffsetReader,
    IJsonlFile,
    IJsonlIndexReader,
    IJsonlSamplePointer,
)
from megatron.energon.flavors.webdataset.structs import FilteredSample
from megatron.energon.source_info import SourceInfo

T_index = TypeVar("T_index", covariant=False)


class IJsonlReader(ABC):
    """
    Class for reading indexed jsonl files containing json samples.

    The common usage patterns and random-access interfaces are provided here.

    Args:
        base_path: The path to the dataset.
        jsonl_path: The path to the jsonl file.
        jsonl_filename: The jsonl file name.
        sample_filter: An optional filter function to select samples by their key.
        index_cache_size: The size of the index cache.
    """

    jsonl_path: EPath
    sample_filter: Optional[Callable[[str], bool]]

    cached_offset_reader: CachedIJsonlOffsetReader
    ijsonl_file: IJsonlFile | None = None

    def __init__(
        self,
        jsonl_path: EPath,
        sample_filter: Optional[Callable[[str], bool]] = None,
        index_cache_size: int = 5,
    ):
        self.jsonl_path = jsonl_path
        self.sample_filter = sample_filter
        self.cached_offset_reader = CachedIJsonlOffsetReader(
            jsonl_path, cache_size=index_cache_size
        )

    def __len__(self) -> int:
        return len(self.cached_offset_reader)

    def __str__(self) -> str:
        return f"IJsonlReader(jsonl_path={self.jsonl_path})"

    def _get_item_by_sample_pointer(
        self,
        sample_pointer: IJsonlSamplePointer,
    ) -> FilteredSample | None:
        """
        Get a sample from the dataset or slice it.

        Args:
            sample_pointer: The sample pointer to get the sample from.
            sample_index: The global index of the sample in the dataset.

        Returns:
            The sample or None if the sample is invalid.
        """

        key = str(sample_pointer.index)
        if self.sample_filter is not None and not self.sample_filter(key):
            return None

        if self.ijsonl_file is None:
            self.ijsonl_file = IJsonlFile(self.jsonl_path.open("rb"))

        json_data = self.ijsonl_file.next(sample_pointer.byte_offset, sample_pointer.byte_size)
        if json_data is None:
            return None

        return FilteredSample(
            __key__=f"{self.jsonl_path.name}/{key}",
            __shard__=self.jsonl_path.name,
            __restore_key__=("Webdataset", sample_pointer.index),
            __sources__=(
                SourceInfo(
                    dataset_path=str(self.jsonl_path),
                    index=sample_pointer.index,
                    shard_name=self.jsonl_path.name,
                    file_names=(f"{key}.json",),
                ),
            ),
            json=json_data,
        )

    def __getitem__(self, idx: int | str) -> FilteredSample | tuple[bytes, SourceInfo] | None:
        """
        Get a sample from the dataset.
        """

        assert isinstance(idx, (int, str)), f"Invalid argument type for __getitem__: {type(idx)}"
        full_entry_name = False
        if isinstance(idx, str):
            if idx.endswith(".json"):
                num_idx = idx.removesuffix(".json")
                full_entry_name = True
            try:
                idx = int(num_idx)
            except ValueError:
                raise ValueError(f"Invalid JSONL sample key: {idx}")

        byte_offset, byte_size = self.cached_offset_reader.get_ijsonl_byte_offset(idx)
        sample: FilteredSample | None = self._get_item_by_sample_pointer(
            IJsonlSamplePointer(
                index=idx,
                byte_offset=byte_offset,
                byte_size=byte_size,
            )
        )

        if sample is None:
            return None

        if full_entry_name:
            assert len(sample["__sources__"]) == 1
            return sample["json"], sample["__sources__"][0]
        else:
            return sample

    def list_all_samples(self) -> Generator[Tuple[str, int, int], None, None]:
        """List all samples in the jsonl file.

        Returns:
            A generator of tuples of (sample_key, size, tar_file_id)
        """
        last_byte_offset = 0
        with IJsonlIndexReader(self.jsonl_path) as ijsonl_index_reader:
            for sample_idx, byte_offset in enumerate(ijsonl_index_reader):
                if last_byte_offset == byte_offset:
                    continue
                yield str(sample_idx), byte_offset - last_byte_offset, 0
                last_byte_offset = byte_offset

    def list_all_sample_parts(self) -> Generator[Tuple[str, int, int], None, None]:
        """List all sample parts in the jsonl file.

        Returns:
            A generator of tuples of (sample_key + "." + part_name, size, tar_file_id)
        """
        last_byte_offset = 0
        with IJsonlIndexReader(self.jsonl_path) as ijsonl_index_reader:
            for sample_idx, byte_offset in enumerate(ijsonl_index_reader):
                if last_byte_offset == byte_offset:
                    continue
                yield f"{sample_idx}.json", byte_offset - last_byte_offset, 0
                last_byte_offset = byte_offset

    def list_sample_parts(self, sample_key: str) -> Generator[Tuple[str, int, int], None, None]:
        """Given a sample key, list all its parts. (E.g. given 1, list 1.jpg, 1.json, etc.)

        Args:
            sample_key: The sample key to list the parts of.

        Returns:
            A generator of tuples of (part_name, size, tar_file_id)
        """
        try:
            sample_idx = int(sample_key)
        except ValueError:
            raise ValueError(f"Invalid JSONL sample key: {sample_key}")

        _, byte_size = self.cached_offset_reader.get_ijsonl_byte_offset(sample_idx)
        yield f"{sample_key}.json", byte_size, 0

    def get_total_size(self) -> int:
        return self.cached_offset_reader.get_total_size()

    def close(self):
        if self.ijsonl_file is not None:
            self.ijsonl_file.close()
        self.cached_offset_reader.close()



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import List, Optional

import torch

from megatron.energon.av import AVDecoder
from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib.epath import EPath
from megatron.energon.errors import warn_deprecated
from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.webdataset import DefaultDecoderWebdatasetFactory


@edataclass
class VidQASample(Sample):
    """Sample type for video question answering."""

    #: The video data containing the image and audio info.
    video: AVDecoder
    #: The context/question for the image.
    context: str
    #: The possible answers. Not set for testing.
    answers: Optional[List[str]] = None
    #: The weights of the possible answers. Optionally available.
    answer_weights: Optional[torch.Tensor] = None


class VidQAWebdataset(DefaultDecoderWebdatasetFactory[VidQASample]):
    __sample_type__ = VidQASample

    def __init__(self, path: EPath, **kwargs):
        warn_deprecated(
            f"{type(self)} is deprecated, use the default instead and set the sample_type:\n"
            f"To convert, update your {path}/.nv-meta/dataset.yaml to:\n"
            f"# remove top-level __module__ and __class__\n"
            f"sample_type:\n"
            f"  __module__: megatron.energon\n"
            f"  __class__: {self.__sample_type__.__name__}\n"
            f"# Keep the remaining content"
        )
        super().__init__(path, **kwargs)
